/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#include <iostream>
#include "StrumpackSparseSolver.hpp"

typedef double scalar;
typedef double real;
//typedef int64_t integer;
typedef int integer;

using namespace strumpack;

int main(int argc, char* argv[]) {
  int n = 30;
  if (argc > 1) n = atoi(argv[1]); // get grid size
  else std::cout << "# please provide grid size" << std::endl;

  StrumpackSparseSolver<scalar,real,integer> spss(argc, argv);
  spss.set_mc64job(0);
  spss.set_matrix_reordering_method(params::GEOMETRIC);
  spss.set_from_options();

  integer N = n * n;
  integer nnz = 5 * N - 4 * n;
  CSRMatrix<scalar,real,integer> A(N, nnz);
  integer* col_ptr = A.get_ptr();
  integer* row_ind = A.get_ind();
  scalar* val = A.get_val();

  nnz = 0;
  col_ptr[0] = 0;
  for (integer row=0; row<n; row++) {
    for (integer col=0; col<n; col++) {
      integer ind = col+n*row;
      val[nnz] = 4.0;
      row_ind[nnz++] = ind;
      if (col > 0)  { val[nnz] = -1.0; row_ind[nnz++] = ind-1; } // left
      if (col < n-1){ val[nnz] = -1.0; row_ind[nnz++] = ind+1; } // right
      if (row > 0)  { val[nnz] = -1.0; row_ind[nnz++] = ind-n; } // up
      if (row < n-1){ val[nnz] = -1.0; row_ind[nnz++] = ind+n; } // down
      col_ptr[ind+1] = nnz;
    }
  }
  A.set_symmetric_sparsity();

  std::vector<scalar> b(N, scalar(1.)), x(N, scalar(0.));

  spss.set_csr_matrix(N, col_ptr, row_ind, val, true);
  spss.reorder(n, n);
  // spss.factor();   // not really necessary, called if needed by solve
  spss.solve(b.data(), x.data());

  // just a check, system is already solved, so solving again
  // with the solution as initial guess should stop immediately
  //spss.solve(b.data(), x.data(), true);

  std::cout << "# COMPONENTWISE SCALED RESIDUAL = " << A.max_scaled_residual(x.data(), b.data()) << std::endl;
  return 0;
}
